import os
import json
import subprocess
import platform

from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtWidgets import QMainWindow, QWidget, QVBoxLayout, QLabel
from PyQt5.QtCore import Qt, QTimer

import serial.tools.list_ports
from singleton import WindowState


# Класс для создания внешнего вида
class UIComponents:
    def __init__(self, parent):
        self.base_dir = os.path.dirname(os.path.abspath(__file__))
        self.logo = os.path.join(self.base_dir, "media/images/logo.svg")
        self.parent = parent
        self.command_listbox = None
        self.delay_entry = None
        self.speed_label = None
        self.angel_label = None
        self.angle_sliders = []
        self.speed_sliders = []
        self.connection_label = None
        self.board_combobox = None
        self.port_combobox = None
        self.message_label = None
        self.available_ports = []
        self.available_board = {}
        self.init_ui()

    def init_ui(self):
        self.parent.setWindowTitle("Менеджер Сервоприводов")
        self.parent.setWindowIcon(QtGui.QIcon(self.logo))
        self.parent.setGeometry(100, 100, 910, 700)
        main_layout = QtWidgets.QVBoxLayout()

        # Создание меню
        self.create_menu(main_layout)

        # Основное содержимое
        layout = QtWidgets.QGridLayout()

        # Сканирование доступных COM-портов
        self.available_ports = self.parent.serial_handler.get_ports()

        # Выбор платы
        board_label = QtWidgets.QLabel("Выберите плату:")
        layout.addWidget(board_label, 0, 0)

        self.board_combobox = QtWidgets.QComboBox()
        self.board_combobox.addItems(self.parent.board_manager.get_board_options())
        layout.addWidget(self.board_combobox, 0, 1)

        # Выбор порта
        port_label = QtWidgets.QLabel("Выберите COM порт:")
        layout.addWidget(port_label, 1, 0)

        self.port_combobox = QtWidgets.QComboBox()
        self.port_combobox.addItems(self.available_ports)
        layout.addWidget(self.port_combobox, 1, 1)

        if self.available_ports:
            self.port_combobox.setCurrentIndex(0)
        else:
            self.port_combobox.addItem("Нет доступных портов")
            self.port_combobox.setCurrentIndex(0)

        connect_button = QtWidgets.QPushButton("Подключиться")
        connect_button.clicked.connect(self.parent.serial_handler.connect_port)
        layout.addWidget(connect_button, 1, 3)

        upload_button = QtWidgets.QPushButton("Загрузить прошивку")
        upload_button.clicked.connect(self.parent.firmware_uploader.upload_firmware)
        layout.addWidget(upload_button, 0, 3)

        refresh_ports_button = QtWidgets.QPushButton("Обновить порты")
        refresh_ports_button.clicked.connect(self.parent.serial_handler.refresh_ports)
        layout.addWidget(refresh_ports_button, 1, 2)

        self.connection_label = QtWidgets.QLabel("Нет соединения")
        layout.addWidget(self.connection_label, 2, 0, 1, 4)

        self.angel_label = QtWidgets.QLabel("Угол поворота")
        layout.addWidget(self.angel_label, 3, 1)

        self.speed_label = QtWidgets.QLabel("Скорость")
        layout.addWidget(self.speed_label, 3, 2)

        # Поле для ввода задержки
        delay_label = QtWidgets.QLabel("Задержка (мс):")
        layout.addWidget(delay_label, 9, 0)

        self.delay_entry = QtWidgets.QLineEdit()
        self.delay_entry.setText("500")
        layout.addWidget(self.delay_entry, 9, 1)

        # Дальнейшие элементы интерфейса (ползунки, кнопки и т.д.)
        self.create_servo_controls(layout)
        self.create_commands_controls(layout)

        main_layout.addLayout(layout)
        self.parent.setLayout(main_layout)

    def create_menu(self, main_layout):
        menubar = QtWidgets.QMenuBar(self.parent)
        help_menu = menubar.addMenu("Справка")

        about_company_action = QtWidgets.QAction("О компании", self.parent)
        about_company_action.triggered.connect(self.parent.show_company_info)
        help_menu.addAction(about_company_action)

        about_program_action = QtWidgets.QAction("О программе", self.parent)
        about_program_action.triggered.connect(self.parent.show_program_info)
        help_menu.addAction(about_program_action)

        main_layout.setMenuBar(menubar)

    def create_servo_controls(self, layout):
        # Создание ползунков для управления углами и скоростью
        servo_names = ["Колонна", "Левое плечо", "Правое плечо", "Захват"]
        self.angle_sliders = []
        self.speed_sliders = []

        for i in range(4):
            servo_label = QtWidgets.QLabel(f"{servo_names[i]}")
            layout.addWidget(servo_label, i + 4, 0)

            # Создание контейнера для ползунка угла и меток
            angle_slider_layout = QtWidgets.QVBoxLayout()

            # Верхние метки значений для углов
            top_labels_layout = QtWidgets.QHBoxLayout()
            top_labels_layout.addWidget(QtWidgets.QLabel("0°"))
            top_labels_layout.addWidget(QtWidgets.QLabel(""))
            top_labels_layout.addWidget(QtWidgets.QLabel(""))
            top_labels_layout.addWidget(QtWidgets.QLabel("60°"))
            top_labels_layout.addWidget(QtWidgets.QLabel(""))
            top_labels_layout.addWidget(QtWidgets.QLabel(""))
            top_labels_layout.addWidget(QtWidgets.QLabel("120°"))
            top_labels_layout.addWidget(QtWidgets.QLabel(""))
            top_labels_layout.addWidget(QtWidgets.QLabel(""))
            top_labels_layout.addWidget(QtWidgets.QLabel("180°"))
            angle_slider_layout.addLayout(top_labels_layout)

            # Создание ползунка угла
            angle_slider = QtWidgets.QSlider(Qt.Horizontal)
            angle_slider.setRange(0, 180)
            angle_slider.setValue(self.parent.servo_handler.target_angles[i])
            angle_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)  # Метки под ползунком
            angle_slider.setTickInterval(60)  # Интервал между метками
            angle_slider.valueChanged.connect(
                lambda val, j=i: self.parent.servo_handler.update_servo(j, val, self.speed_sliders[j].value())
            )
            angle_slider_layout.addWidget(angle_slider)

            layout.addLayout(angle_slider_layout, i + 4, 1)

            # Создание контейнера для ползунка скорости и меток
            speed_slider_layout = QtWidgets.QVBoxLayout()

            # Верхние метки значений для скорости
            speed_top_labels_layout = QtWidgets.QHBoxLayout()
            speed_top_labels_layout.addWidget(QtWidgets.QLabel("1°/с"))
            speed_top_labels_layout.addWidget(QtWidgets.QLabel(""))
            speed_top_labels_layout.addWidget(QtWidgets.QLabel("25°/с"))
            speed_top_labels_layout.addWidget(QtWidgets.QLabel(""))
            speed_top_labels_layout.addWidget(QtWidgets.QLabel("50°/с"))
            speed_top_labels_layout.addWidget(QtWidgets.QLabel(""))
            speed_top_labels_layout.addWidget(QtWidgets.QLabel("75°/с"))
            speed_top_labels_layout.addWidget(QtWidgets.QLabel(""))
            speed_top_labels_layout.addWidget(QtWidgets.QLabel("100°/с"))
            speed_slider_layout.addLayout(speed_top_labels_layout)

            # Создание ползунка скорости
            speed_slider = QtWidgets.QSlider(Qt.Horizontal)
            speed_slider.setRange(1, 100)
            speed_slider.setValue(self.parent.servo_handler.target_speeds[i])
            speed_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)  # Метки под ползунком
            speed_slider.setTickInterval(25)  # Интервал между метками
            speed_slider.valueChanged.connect(
                lambda val, j=i: self.parent.servo_handler.update_servo(j, self.angle_sliders[j].value(), val)
            )
            speed_slider_layout.addWidget(speed_slider)

            layout.addLayout(speed_slider_layout, i + 4, 2)

            self.angle_sliders.append(angle_slider)
            self.speed_sliders.append(speed_slider)

    def create_commands_controls(self, layout):
        # Кнопки для управления командами
        add_button = QtWidgets.QPushButton("Добавить команду")
        add_button.clicked.connect(self.parent.command_manager.add_command)
        layout.addWidget(add_button, 10, 0)

        delete_button = QtWidgets.QPushButton("Удалить команду")
        delete_button.clicked.connect(self.parent.command_manager.delete_command)
        layout.addWidget(delete_button, 10, 1)

        self.command_listbox = QtWidgets.QListWidget()
        layout.addWidget(self.command_listbox, 11, 0, 1, 2)

        step_play_button = QtWidgets.QPushButton("Пошаговое воспроизведение")
        step_play_button.clicked.connect(self.parent.command_manager.step_play)
        layout.addWidget(step_play_button, 10, 2)

        auto_play_button = QtWidgets.QPushButton("Автоматическое воспроизведение")
        auto_play_button.clicked.connect(self.parent.command_manager.auto_play)
        layout.addWidget(auto_play_button, 10, 3)

        # Кнопки для сохранения и загрузки команд
        save_button = QtWidgets.QPushButton("Сохранить команды в файл")
        save_button.clicked.connect(self.parent.command_manager.save_command)
        layout.addWidget(save_button, 4, 3)

        load_button = QtWidgets.QPushButton("Загрузить команды из файла")
        load_button.clicked.connect(self.parent.command_manager.load_command)
        layout.addWidget(load_button, 5, 3)

        # Метка для вывода сообщений пользователю
        self.message_label = QtWidgets.QLabel("")
        layout.addWidget(self.message_label, 11, 2, 1, 3)


# Класс для подключения к порту
class SerialHandler:
    def __init__(self, parent):
        self.parent = parent
        self.ser = None

    @staticmethod
    def get_ports():
        return [port.device for port in serial.tools.list_ports.comports()]

    def refresh_ports(self):
        self.parent.ui.available_ports = self.get_ports()
        self.parent.ui.port_combobox.clear()
        self.parent.ui.port_combobox.addItems(self.parent.ui.available_ports)
        if self.parent.ui.available_ports:
            self.parent.ui.port_combobox.setCurrentIndex(0)
        else:
            self.parent.ui.port_combobox.addItem("Нет доступных портов")
            self.parent.ui.port_combobox.setCurrentIndex(0)
        self.parent.ui.connection_label.setText("Список портов обновлен")

    def connect_port(self):
        port = self.parent.ui.port_combobox.currentText()
        if port and port != "Нет доступных портов":
            try:
                self.ser = serial.Serial(port, 9600, timeout=1)
                self.parent.ui.connection_label.setText(f"Подключено к {port}")
            except serial.SerialException:
                self.parent.ui.connection_label.setText("Порт не доступен")
        else:
            self.parent.ui.connection_label.setText("Порт не выбран")

    def disconnect_port(self):
        """Отключение от порта"""
        if self.ser and self.ser.is_open:
            self.ser.close()
            self.parent.ui.connection_label.setText("Порт отключен")

    def reconnect_port(self, port):
        """Переподключение к порту"""
        try:
            self.ser = serial.Serial(port, 9600, timeout=1)
            self.parent.ui.connection_label.setText(f"Подключено к {port}")
        except serial.SerialException:
            self.parent.ui.connection_label.setText("Не удалось переподключиться к порту")

# Класс для работы над серводвигателями
class ServoHandler:
    def __init__(self, parent):
        self.parent = parent
        self.target_angles = [90, 90, 90, 90]
        self.target_speeds = [50, 50, 50, 50]
        self.command_queue = []
        self.timer = QTimer(self.parent)
        self.timer.timeout.connect(self.process_command_queue)
        self.timer.start(1)

    def update_servo(self, num, angle_value, speed_value):
        self.target_angles[num] = int(angle_value)
        self.target_speeds[num] = int(speed_value)
        self.send_servo_angle(num + 1, self.target_angles[num], self.target_speeds[num])

    def send_servo_angle(self, servo_num, angle, speed):
        if self.parent.serial_handler.ser is None:
            return QtWidgets.QMessageBox.critical(self.parent, "Предупреждение", "Соединение с портом не установлено.")
        if not self.parent.serial_handler.ser.is_open:
            QtWidgets.QMessageBox.critical(self.parent, "Предупреждение", "Порт закрыт для передачи сигнала.")
        if self.parent.serial_handler.ser and self.parent.serial_handler.ser.is_open:
            command = f"{servo_num},{angle},{speed}\n"
            self.command_queue.append(command)

    def process_command_queue(self):
        if self.parent.serial_handler.ser and self.parent.serial_handler.ser.is_open and len(self.command_queue) > 0:
            command = self.command_queue.pop(0)
            try:
                self.parent.serial_handler.ser.write(command.encode("utf-8"))
            except Exception as e:
                QtWidgets.QMessageBox.critical(self.parent, "Ошибка", f"Произошла ошибка при отправке данных: {str(e)}")


# Класс для работы с CRUD командами
class CommandManager:
    def __init__(self, parent):
        self.parent = parent
        self.command_list = []
        self.play_index = 0
        self.auto_play_timer = QTimer(self.parent)
        self.auto_play_timer.timeout.connect(self.play_next_command)

    def add_command(self):
        command = {
            "angles": self.parent.servo_handler.target_angles.copy(),
            "speeds": self.parent.servo_handler.target_speeds.copy()
        }
        self.command_list.append(command)
        self.update_command_listbox()

    def delete_command(self):
        selected_index = self.parent.ui.command_listbox.currentRow()
        if selected_index >= 0:
            self.command_list.pop(selected_index)
            self.update_command_listbox()

    def update_command_listbox(self):
        self.parent.ui.command_listbox.clear()
        for num, cmd in enumerate(self.command_list):
            self.parent.ui.command_listbox.addItem(f"Команда {num + 1}: {cmd}")

    def step_play(self):
        selected_indices = self.parent.ui.command_listbox.selectedIndexes()
        if selected_indices:
            current_index = selected_indices[0].row()
            self.play_command(current_index)
            self.parent.ui.message_label.setText(f"Воспроизведена команда {current_index + 1}")

    def play_command(self, num):
        if num < len(self.command_list):
            angles = self.command_list[num]["angles"]
            speeds = self.command_list[num]["speeds"]
            for i in range(4):
                self.parent.servo_handler.send_servo_angle(i + 1, angles[i], speeds[i])

    def auto_play(self):
        delay = int(self.parent.ui.delay_entry.text())
        if self.command_list:
            self.play_index = 0
            self.auto_play_timer.start(delay)

    def play_next_command(self):
        if self.play_index < len(self.command_list):
            self.play_command(self.play_index)
            self.parent.ui.message_label.setText(f"Воспроизведена команда {self.play_index + 1}")
            self.play_index += 1
        else:
            self.auto_play_timer.stop()

    def save_command(self):
        if not os.path.exists("commands"):
            os.makedirs("commands")
            
        with open("commands/commands.json", "w") as f:
            json.dump(self.command_list, f)
        self.parent.ui.message_label.setText("Команды сохранены в commands.json")

    def load_command(self):
        if not os.path.exists("commands"):
            os.makedirs("commands")
        try:
            with open("commands/commands.json", "r") as f:
                self.command_list = json.load(f)
            self.update_command_listbox()
            self.parent.ui.message_label.setText("Команды загружены из commands.json")
        except FileNotFoundError:
            self.command_list = []
            self.parent.ui.message_label.setText("Файл команд не найден.")


# Класс для загрузки прошивки на плату
class FirmwareUploader:
    def __init__(self, parent):
        self.upload_finished = QtCore.pyqtSignal(str)
        self.parent = parent
        self.arduino_code = """\
#include <Servo.h>

Servo column;
Servo left_shoulder;
Servo right_shoulder;
Servo grip;

void setup() {
  Serial.begin(9600);
  column.attach(4);
  left_shoulder.attach(5);
  right_shoulder.attach(6);
  grip.attach(7);
}

void loop() {
  if (Serial.available() > 0) {
    String input = Serial.readStringUntil('\\n');
    int commaIndex = input.indexOf(',');
    if (commaIndex > 0) {
      int servoNum = input.substring(0, commaIndex).toInt();
      int angle = input.substring(commaIndex + 1).toInt();

      switch (servoNum) {
        case 1:
          column.write(angle);
          break;
        case 2:
          left_shoulder.write(angle);
          break;
        case 3:
          right_shoulder.write(angle);
          break;
        case 4:
          grip.write(angle);
          break;
      }
    }
  }
}
"""

    @staticmethod
    def create_ino_file(file_name, content):
        """
        Создает .ino файл с указанным содержимым.

        :param file_name: Имя файла (должно оканчиваться на .ino).
        :param content: Содержимое файла.
        :return: Полный путь к созданному файлу.
        """
        if not file_name.endswith('.ino'):
            raise ValueError("Имя файла должно заканчиваться на '.ino'")

        with open(file_name, 'w') as file:
            file.write(content)

        return os.path.abspath(file_name)

    @staticmethod
    def is_arduino_port(port):
        result = subprocess.run(["arduino-cli", "board", "list"], capture_output=True, text=True)
        return port in result.stdout

    def upload_firmware(self):
        current_directory_name = f"{os.path.basename(os.getcwd())}.ino"
        temp_name = self.create_ino_file(current_directory_name, self.arduino_code)
        port = self.parent.ui.port_combobox.currentText()
        board_key = self.parent.ui.board_combobox.currentText()
        board = self.parent.board_manager.available_board[board_key]

        if not os.path.exists(temp_name):
            QtWidgets.QMessageBox.critical(self.parent, "Ошибка", f"Файл прошивки не найден:\n{temp_name}")
            return

        if not port or port == "Нет доступных портов":
            QtWidgets.QMessageBox.critical(self.parent, "Ошибка", "Выберите COM порт.")
            return

        if not self.is_arduino_port(port):
            QtWidgets.QMessageBox.critical(self.parent, "Ошибка", "Выбранный порт не поддерживает Arduino плату.")
            return

        self.parent.serial_handler.disconnect_port()

        try:
            if platform.system() == "Windows":
                subprocess.run(["arduino-cli", "core", "install", "arduino:avr"], check=True,
                               creationflags=subprocess.CREATE_NO_WINDOW)
                subprocess.run(["arduino-cli", "lib", "install", "Servo"], check=True,
                               creationflags=subprocess.CREATE_NO_WINDOW)
            else:
                subprocess.run(["arduino-cli", "core", "install", "arduino:avr"], check=True)
                subprocess.run(["arduino-cli", "lib", "install", "Servo"], check=True)

            compile_command = [
                "arduino-cli",
                "compile",
                "--fqbn", board,
                temp_name
            ]

            if platform.system() == "Windows":
                subprocess.run(compile_command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               creationflags=subprocess.CREATE_NO_WINDOW)
            else:
                subprocess.run(compile_command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

            upload_command = [
                "arduino-cli",
                "upload",
                "-p", port,
                "--fqbn", board,
                temp_name
            ]
            if platform.system() == "Windows":
                process = subprocess.run(upload_command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                         creationflags=subprocess.CREATE_NO_WINDOW, timeout=20)
            else:
                process = subprocess.run(upload_command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=20)

            if process.returncode == 0:
                QtWidgets.QMessageBox.information(self.parent, "Успех", "Прошивка успешно загружена!")
            if process.returncode != 0:
                error_message = stderr.decode().strip()
                if "avrdude: stk500_recv(): programmer is not responding" in error_message:
                    QtWidgets.QMessageBox.critical(self.parent, "Ошибка",
                                                   "Программатор не отвечает. Проверьте плату и порт.")
                elif "not in sync" in error_message:
                    QtWidgets.QMessageBox.critical(self.parent, "Ошибка",
                                                   "Ошибка: не удалось синхронизироваться. Убедитесь, что плата подключена и выбрана правильно.")

                elif "stk500v2_ReceiveMessage(): timeout" in error_message:
                    QtWidgets.QMessageBox.critical(self.parent, "Ошибка",
                                                   "Ошибка тайм-аута. Убедитесь, что плата подключена и выбрана правильная конфигурация.")
                else:
                    QtWidgets.QMessageBox.critical(self.parent, "Ошибка", f"Ошибка загрузки: {error_message}")

        except subprocess.TimeoutExpired:
            # Если загрузка длится слишком долго — выводим сообщение и прерываем процесс
            QtWidgets.QMessageBox.critical(self.parent, "Предупреждение", "Убедитесь, что плата подключена и выбрана правильно.")
        except subprocess.CalledProcessError as e:
            error_message = e.stderr.decode(errors="replace").strip()
            QtWidgets.QMessageBox.critical(self.parent, "Ошибка",
                                           f"Не удалось загрузить прошивку:\n{error_message}")
        except Exception as e:
            QtWidgets.QMessageBox.critical(self.parent, "Неожиданная ошибка", str(e))

        finally:
            self.parent.serial_handler.reconnect_port(port)


class BoardManager:
    def __init__(self):
        self.available_board = {}  # Создаём атрибут здесь
        base_dir = os.path.dirname(os.path.abspath(__file__))
        json_path = os.path.join(base_dir, "boards.json")
        self.load_boards_from_json(json_path)

    def load_boards_from_json(self, filename):
        if not os.path.exists(filename):
            print(f"Файл {filename} не найден!")
            return  # Не вызываем ошибку, просто пропускаем
        with open(filename, "r", encoding="utf-8") as file:
            self.available_board = json.load(file)

    def get_board_options(self):
        return list(self.available_board.keys())


# Класс для создания приложения
class ServoControllerApp(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self.serial_handler = SerialHandler(self)
        self.servo_handler = ServoHandler(self)
        self.command_manager = CommandManager(self)
        self.firmware_uploader = FirmwareUploader(self)
        self.board_manager = BoardManager()
        self.ui = UIComponents(self)

    def show_company_info(self):
        QtWidgets.QMessageBox.information(self, "О компании", "Информация о компании.")

    def show_program_info(self):
        QtWidgets.QMessageBox.information(self, "О программе", "Информация о программе.")

    def closeEvent(self, event):
        window_state = WindowState()
        window_state.set_window_instance(None)  # Сбрасываем ссылку на окно при закрытии
        event.accept()


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    window = ServoControllerApp()
    window.show()
    app.exec_()
